from .model import Model, APIResponse
from .category import Category
from .product import Product, MediaEntry, ProductAttribute
from .order import Order, OrderItem
from .invoice import Invoice, InvoiceItem



from __future__ import annotations
from functools import cached_property
from abc import ABC, abstractmethod
from typing import TYPE_CHECKING, Union, Optional, List
from magento import clients
import urllib.parse
import inspect

if TYPE_CHECKING:
    from magento.search import SearchQuery


class Model(ABC):

    """The abstract base class of all API response wrapper classes

    **Overview**

    * A :class:`Model` wraps the response ``data`` from an API ``endpoint``
    * Several endpoints have subclasses with additional methods to retrieve/update data
    * All other endpoints are wrapped using a general :class:`~.APIResponse`
    * The endpoint's corresponding :class:`~.SearchQuery` can be accessed via :meth:`~.query_endpoint`
    """

    DOCUMENTATION: str = None  #: Link to the Official Magento 2 API documentation for the endpoint wrapped by the Model
    IDENTIFIER: str = None  #: The API response field that the endpoint's :attr:`~.Model.uid` comes from

    def __init__(self, data: dict, client: clients.Client, endpoint: str, private_keys: bool = True):
        """Initialize a :class:`Model` object from an API response and the ``endpoint`` that it came from

        ...

        .. tip:: The ``endpoint`` is used to:

           * Generate the :meth:`~.url_for` any requests made by subclass-specific methods
           * Match the :class:`~.Model` to its corresponding :class:`~.SearchQuery` object
           * Determine how to :meth:`~.Model.parse` new :class:`~.Model` objects from API responses

        ...

        :param data: the JSON from an API response to use as source data
        :param client: an initialized :class:`~.Client`
        :param endpoint: the API endpoint that the :class:`Model` wraps
        :param private_keys: if ``True``, sets the keys in the :attr:`~.excluded_keys` as private attributes
            (prefixed with ``__``) instead of fully excluding them

        """
        if not isinstance(data, dict):
            raise TypeError(f'Parameter "data" must be of type {dict}')
        if not isinstance(endpoint, str):
            raise TypeError(f'Parameter "endpoint" must be of type {str}')
        if not isinstance(client, clients.Client):
            raise TypeError(f'Parameter "client" must be of type {clients.Client}')

        self.data = data
        self.client = client
        self.endpoint = endpoint
        self.logger = client.logger
        self.set_attrs(data, private_keys=private_keys)

    def set_attrs(self, data: dict, private_keys: bool = True) -> None:
        """Initializes object attributes using the JSON from an API response as the data source

        Called at the time of object initialization, but can also be used to update the source data and
        reinitialize the attributes without creating a new object

        :param data: the API response JSON to use as the object source data
        :param private_keys: if set to True, will set the :attr:`~.excluded_keys` as private attributes
            (prefixed with ``__``) instead of fully excluding them

        .. admonition:: **Private Keys Clarification**
           :class: info

           Let's say that ``"status"`` is in the :attr:`~.excluded_keys`

           * No matter what, the ``status`` attribute will not be set on the :class:`Model`
           * If ``private_keys==True``, the ``__status`` attribute will be set (using the ``status`` data)
           * If ``private_keys==False``, the data from ``status`` is completely excluded
        """
        keys = set(data) - set(self.excluded_keys)
        for key in keys:
            if key == 'custom_attributes':
                if attrs := data[key]:
                    setattr(self, key, self.unpack_attributes(attrs))
            else:
                setattr(self, key, data[key])

        if private_keys:
            private = '_' + self.__class__.__name__ + '__'
            for key in self.excluded_keys:
                setattr(self, private + key, data.get(key))

        self.data = data

    @property
    @abstractmethod
    def excluded_keys(self) -> List[str]:
        """API response keys that shouldn't be set as object attributes by :meth:`~.set_attrs`

        :returns: list of API response keys that shouldn't be set as attributes
        """
        pass

    @property
    def uid(self) -> Union[str, int]:
        """Unique item identifier; used in the url of the :meth:`~.Model.data_endpoint`"""
        return self.data.get(self.IDENTIFIER)

    def data_endpoint(self, scope: Optional[str] = None) -> str:
        """Endpoint to use when requesting/updating the item's data

        :param scope: the scope to generate the :meth:`~.url_for`
        """
        return self.client.url_for(f'{self.endpoint}/{self.uid}', scope)

    def query_endpoint(self) -> SearchQuery:
        """Initializes and returns the :class:`~.SearchQuery` object corresponding to the Model's ``endpoint``

        :returns: a :class:`~.SearchQuery` or subclass, depending on the ``endpoint``
        """
        return self.client.search(self.endpoint)

    def parse(self, response: dict) -> Model:
        """Uses the instance's corresponding :class:`~.SearchQuery` to parse an API response

        :param response: API response dict to use as source data
        :returns: a :class:`~.Model` with the same ``endpoint`` as the calling instance
        """
        return self.query_endpoint().parse(response)

    def refresh(self, scope: Optional[str] = None) -> bool:
        """Updates object attributes in place using current data from the :meth:`~.data_endpoint`

        .. hint:: :meth:`~.refresh` can be used to switch the scope of the source data
           without creating a new object or changing the :attr:`.Client.scope`

           .. admonition:: Example
              :class: example

              ::

                # Get product data on 'default' scope
                >>> product = client.products.by_sku('sku42')
                # Get fresh product data from different scope
                >>> product.refresh('all')

                Refreshed <Magento Product: sku42> on scope all

        :param scope: the scope to send the request on; uses the :attr:`.Client.scope` if not provided
        """
        url = self.data_endpoint(scope)
        response = self.client.get(url)

        if response.ok:
            self.clear(*self.cached)
            self.set_attrs(response.json())
            self.logger.info(
                f"Refreshed {self} on scope {self.get_scope_name(scope)}"
            )
            return True
        else:
            self.logger.error(  # Actual error message is logged by client
                f"Failed to refresh {self} on scope {self.get_scope_name(scope)}"
            )
            return False

    @staticmethod
    def unpack_attributes(attributes: List[dict], key: str = 'attribute_code') -> dict:
        """Unpacks a list of attribute dictionaries into a single dictionary

        .. admonition:: Example
           :class: example

           ::

            >> custom_attrs = [{'attribute_code': 'attr', 'value': 'val'},{'attribute_code': 'will_to_live', 'value': '0'}]
            >> print(Model.unpack_attributes(custom_attrs))

            {'attr': 'val', 'will_to_live': '0'}

        :param attributes: a list of custom attribute dictionaries
        :param key: the key used in the attribute dictionary (ex. ``attribute_code`` or ``label``)
        :returns: a single dictionary of all custom attributes formatted as ``{"attr": "val"}``
        """
        return {attr[key]: attr['value'] for attr in attributes}

    @staticmethod
    def pack_attributes(attribute_data: dict, key: str = 'attribute_code') -> List[dict]:
        """Packs a dictionary containing attributes into a list of attribute dictionaries

        .. admonition:: **Example**
           :class: example

           ::

            >> attribute_data = {'special_price': 12, 'meta_title': 'My Product'}
            >> print(Model.pack_attributes(attribute_data))
            >> print(Model.pack_attributes(attribute_data, key='label'))

            [{'attribute_code': 'special_price', 'value': 12}, {'attribute_code': 'meta_title', 'value': 'My Product'}]
            [{'label': 'special_price', 'value': 12}, {'label': 'meta_title', 'value': 'My Product'}]


        :param attribute_data: a dictionary containing attribute data
        :param key: the key to use when packing the attributes (ex. ``attribute_code`` or ``label``)
        :returns: a list of dictionaries formatted as ``{key : "attr", "value": "value"}``
        """
        return [{key: attr, "value": val} for attr, val in attribute_data.items()]

    @staticmethod
    def encode(string: str) -> str:
        """URL-encode with :mod:`urllib.parse`; used for requests that could contain special characters

        :param string: the string to URL-encode
        """
        if urllib.parse.unquote_plus(string) != string:
            return string  # Already encoded
        return urllib.parse.quote_plus(string)

    @cached_property
    def cached(self) -> List[str]:
        """Names of properties that are wrapped with :func:`functools.cached_property`"""
        return [member for member, val in inspect.getmembers(self.__class__) if
                isinstance(val, cached_property) and member != 'cached']

    def clear(self, *keys: str) -> None:
        """Deletes the provided keys from the object's :attr:`__dict__`

        To clear all cached properties::

            >> self.clear(*self.cached)

        :param keys: name of the object attribute(s) to delete
        """
        for key in keys:
            self.__dict__.pop(key, None)
        self.logger.debug(f'Cleared {keys} from {self}')

    def get_scope_name(self, scope: str) -> str:
        """Returns the appropriate scope name to use for logging messages"""
        return scope or 'default' if scope is not None else self.client.scope or 'default'


class APIResponse(Model):

    IDENTIFIER = 'entity_id'  # Most endpoints use this field

    def __init__(self, data: dict, client: clients.Client, endpoint: str):
        """A generic :class:`Model` subclass

        Wraps API responses when there isn't a :class:`Model` subclass defined for the ``endpoint``

        :param data: the API response from an API endpoint
        :param client: an initialized :class:`~.Client` object
        :param endpoint: the endpoint that the API response came from
        """
        super().__init__(
            data=data,
            client=client,
            endpoint=endpoint,
            private_keys=False
        )

    @property
    def excluded_keys(self) -> List[str]:
        return []

    @property
    def uid(self) -> Optional[int]:
        """Unique item identifier

        .. note:: Since the :class:`~.APIResponse` can wrap any endpoint, the response
           is checked for commonly used id fields (``entity_id`` and ``id``)

           If the endpoint doesn't use those fields, ``None`` will be returned
        """
        if not (uid := super().uid):
            uid = self.data.get('id', None)
        return uid

    def data_endpoint(self, scope: Optional[str] = None) -> Optional[str]:
        if self.uid:
            return super().data_endpoint(scope)
        else:
            self.logger.info(
                f'Unable to determine uid field for API response from "{self.endpoint}"')



from __future__ import annotations
from functools import cached_property
from typing import Union, TYPE_CHECKING, Optional, List
from . import Model

if TYPE_CHECKING:
    from magento import Client
    from magento.search import SearchQuery
    from . import Category, Order, OrderItem, Invoice


class Product(Model):

    """Wrapper for the ``products`` endpoint"""

    STATUS_ENABLED = 1
    STATUS_DISABLED = 2

    VISIBILITY_NOT_VISIBLE = 1
    VISIBILITY_CATALOG = 2
    VISIBILITY_SEARCH = 3
    VISIBILITY_BOTH = 4

    DOCUMENTATION = 'https://adobe-commerce.redoc.ly/2.3.7-admin/tag/products/'
    IDENTIFIER = 'sku'

    def __init__(self, data: dict, client: Client):
        """Initialize a Product object using an API response from the ``products`` endpoint

        :param data: the API response from the ``products`` endpoint
        :param client: an initialized :class:`~.Client` object
        """
        super().__init__(
            data=data,
            client=client,
            endpoint='products',
            private_keys=True
        )

    def __repr__(self):
        return f'<Magento Product: {self.sku}>'

    @property
    def excluded_keys(self):
        return ['media_gallery_entries']

    @property
    def uid(self) -> Union[str, int]:
        return self.encoded_sku

    def update_stock(self, qty: int) -> bool:
        """Updates the stock quantity

        :param qty: the new stock quantity
        """
        url = f'{self.data_endpoint()}/stockItems/{self.stock_item_id}'
        payload = {
            "stock_item": {
                "qty": qty,
                "is_in_stock": qty > 0
            },
            'save_options': True
        }
        response = self.client.put(url, payload)

        if response.ok:
            self.refresh()
            self.logger.info(f'Updated stock to {self.stock} for {self}')
            return True

        else:
            self.logger.error(
                f'Failed to update stock for {self} with status code {response.status_code}' + '\n' +
                f'Message: {response.json()}'
            )
            return False

    def update_status(self, status: int) -> bool:
        """Update the product status

        :param status: either 1 (for :attr:`~.STATUS_ENABLED`) or 2 (for :attr:`~.STATUS_DISABLED`)
        """
        if status not in [Product.STATUS_ENABLED, Product.STATUS_DISABLED]:
            raise ValueError('Invalid status provided')

        return self.update_attributes({'status': status})

    def update_price(self, price: Union[int, float]) -> bool:
        """Update the product price

        :param price: the new price
        """
        return self.update_attributes({'price': price})

    def update_special_price(self, price: Union[float, int]) -> bool:
        """Update the product special price

        :param price: the new special price
        """
        if price < self.price:
            return self.update_custom_attributes({'special_price': price})

        self.logger.error(f'Sale price for {self} must be less than current price ({self.price})')
        return False

    def update_name(self, name: str, scope: Optional[str] = None) -> bool:
        """Update the product name

        :param name: the new name to use
        :param scope: the scope to send the request on; will use the :attr:`.Client.scope` if not provided
        """
        return self.update_attributes({'name': name}, scope)

    def update_description(self, description: str, scope: Optional[str] = None) -> bool:
        """Update the product description

        :param description: the new HTML description to use
        :param scope: the scope to send the request on; will use the :attr:`.Client.scope` if not provided
        """
        return self.update_custom_attributes({'description': description}, scope)

    def update_metadata(self, metadata: dict, scope: Optional[str] = None) -> bool:
        """Update the product metadata

        :param metadata: the new ``meta_title``, ``meta_keyword`` and/or ``meta_description`` to use
        :param scope: the scope to send the request on; will use the :attr:`.Client.scope` if not provided
        """
        attributes = {k: v for k, v in metadata.items() if k in ('meta_title', 'meta_keyword', 'meta_description')}
        return self.update_custom_attributes(attributes, scope)

    def update_attributes(self, attribute_data: dict, scope: Optional[str] = None) -> bool:
        """Update top level product attributes with scoping taken into account

        .. note:: Product attributes can have a ``Global``, ``Store View`` or ``Website`` scope

            :Global Attributes:
                Values are updated on all store views and the admin
            :Website Attributes:
                Values are updated on all store views
            :Store View Attributes:
                Values are updated on the store view specified in the request ``scope``

        A second request will be made to update ``Store View`` and ``Website`` attributes on the admin,
        depending on how many :class:`~.Store` :attr:`~.views` you have:

        * **1 View:** admin values are updated for all attributes, regardless of scope
        * **2+ Views:** admin values are updated only for :attr:`~.website_product_attributes`

        :param attribute_data: a dictionary of product attributes to update
        :param scope: the scope to send the request on; will use the :attr:`.Client.scope` if not provided
        """
        if self.client.store.is_single_store:
            return self._update_single_store(attribute_data)

        if not self._update_attributes(attribute_data, scope):
            return False

        if website_attrs := self.client.store.filter_website_attrs(attribute_data):
            return self._update_attributes(website_attrs, scope='all')
        return True

    def update_custom_attributes(self, attribute_data: dict, scope: Optional[str] = None) -> bool:
        """Update custom attributes with scoping taken into account

        See :meth:`~update_attributes` for details

        .. admonition:: Important
           :class: important-af

           This method only supports updating **custom attributes**

        :param attribute_data: a dictionary of custom attributes to update
        :param scope: the scope to send the request on; will use the :attr:`.Client.scope` if not provided
        """
        attributes = {'custom_attributes': self.pack_attributes(attribute_data)}

        if self.client.store.is_single_store:
            return self._update_single_store(attributes)

        if not self._update_attributes(attributes, scope):
            return False

        if website_attributes := self.client.store.filter_website_attrs(attribute_data):
            return self._update_attributes({'custom_attributes': self.pack_attributes(website_attributes)}, scope='all')
        return True

    def _update_single_store(self, attribute_data: dict) -> bool:
        """Internal function for updating a store with a single store view

        All attributes will be updated on the ``default`` and ``all`` scope,
        ensuring that the frontend and admin always have the same product data

        :param attribute_data: a dictionary of custom product attributes to update
        """
        for store_code in (None, 'all'):
            if not self._update_attributes(attribute_data, store_code):
                return False  # Avoid updating admin if store update fails

        self.refresh()  # Back to default scope
        return True

    def _update_attributes(self, attribute_data: dict, scope: Optional[str] = None) -> bool:
        """Sends a PUT request to update **top-level** product attributes

        .. tip:: to update attributes or custom attributes with attribute scope taken into account,
            use :meth:`~.update_attributes` or :meth:`~.update_custom_attributes` instead

        :param attribute_data: dict containing any number of top-level attributes to update
        :param scope: the scope to send the request on; will use the :attr:`.Client.scope` if not provided
        """
        url = self.data_endpoint(scope)
        payload = {
            "product": {
                "sku": self.sku
            },
            'save_options': True
        }
        payload['product'].update(attribute_data)

        response = self.client.put(url, payload)
        if response.ok:
            self.refresh(scope)
            for key in attribute_data:
                self.logger.info(
                    f"Updated {key} for {self} to {getattr(self, key)} on scope {self.get_scope_name(scope)}")
            return True
        else:
            self.logger.error(
                f'Failed with status code {response.status_code}' + '\n' +
                f'Message: {response.json()}')
            return False

    def get_orders(self) -> Optional[Order | List[Order]]:
        """Searches for orders that contain the product

        If the product is configurable, returns orders containing any of its child products

        :returns: orders that contain the product, as an individual or list of :class:`~.Order` objects
        """
        return self.client.orders.by_product(self)

    def get_order_items(self) -> Optional[OrderItem | List[OrderItem]]:
        """Searches for order items that contain the product

        If the product is configurable, returns order items containing any of its child products

        :returns: order items that contain the product, as an individual or list of :class:`~.OrderItem` objects
        """
        return self.client.order_items.by_product(self)

    def get_invoices(self) -> Optional[Invoice | List[Invoice]]:
        """Searches for invoices that contain the product

        If the product is configurable, returns invoices containing any of its child products

        :returns: invoices that contain the product, as an individual or list of :class:`~.Invoice` objects
        """
        return self.client.invoices.by_product(self)

    def delete(self) -> bool:
        """Deletes the product

        .. hint:: If you delete a product by accident, the :class:`Product` object's ``data``
         attribute will still contain the raw data, which can be used to recover it.

         Alternatively, don't delete it by accident.
        """
        url = self.data_endpoint()
        response = self.client.delete(url)

        if response.ok and response.json() is True:
            self.logger.info(f'Deleted {self}')
            return True
        else:
            self.logger.error(
                f'Failed to delete {self}. Message: {response.json()["message"]}'
            )
            return False

    def get_children(self, refresh: bool = False, scope: Optional[str] = None) -> List[Product]:
        """Retrieve the child simple products of a configurable product

        :param refresh: if True, calls :meth:`~.Model.refresh` on the child products to retrieve full data
        :param scope: the scope to refresh the children on (when ``refresh=True``)
        """
        if refresh:
            for child in self.children:
                child.refresh(scope)
        return self.children

    @cached_property
    def children(self) -> List[Product]:
        """If the Product is a configurable product, returns a list of its child products"""
        if self.type_id == 'configurable':
            url = self.client.url_for(f'configurable-products/{self.encoded_sku}/children')
            if (response := self.client.get(url)).ok:
                return [self.parse(child) for child in response.json()]
            else:
                self.logger.error(f'Failed to get child products of {self}')
        else:
            self.logger.info('Only configurable products have child SKUs')
        return []

    @cached_property
    def categories(self) -> Optional[Category | List[Category]]:
        """Categories the product is in, returned as a list of :class:`~.Category` objects"""
        category_ids = self.custom_attributes.get('category_ids', [])
        return self.client.categories.by_list('entity_id', category_ids)

    @cached_property
    def media_gallery_entries(self) -> List[MediaEntry]:
        """The product's media gallery entries, returned as a list of :class:`MediaEntry` objects"""
        return [MediaEntry(self, entry) for entry in self.__media_gallery_entries]

    @cached_property
    def thumbnail(self) -> MediaEntry:
        """The :class:`MediaEntry` corresponding to the product's thumbnail"""
        for entry in self.media_gallery_entries:
            if entry.is_thumbnail:
                return entry

    @property
    def thumbnail_link(self) -> str:
        """Link of the product's :attr:`~.thumbnail` image"""
        if self.thumbnail:
            return self.thumbnail.link

    def get_media_by_id(self, entry_id: int) -> MediaEntry:
        """Access a :class:`MediaEntry` of the product by id

        :param entry_id: the id of the media gallery entry
        """
        for entry in self.media_gallery_entries:
            if entry.id == entry_id:
                return entry

    @property
    def encoded_sku(self) -> str:
        """URL-encoded SKU, which is used in request endpoints"""
        return self.encode(self.sku)

    @cached_property
    def option_skus(self) -> List[str]:
        """The full SKUs for the product's customizable options, if they exist

        .. hint:: When a product with customizable options is ordered, these SKUs are used by the API when
            retrieving and searching for :class:`~.Order` and :class:`~.OrderItem` data
        """
        option_skus = []
        if hasattr(self, 'options'):
            for option in self.options:
                base_sku = option['product_sku']
                for val in option['values']:
                    if val.get('sku'):
                        option_skus.append(base_sku + '-' + val['sku'])
        return option_skus

    @property
    def stock(self) -> int:
        """Current stock quantity"""
        if self.stock_item:
            return self.stock_item['qty']

    @property
    def stock_item(self) -> dict:
        """Stock data from the StockItem Interface"""
        if hasattr(self, 'extension_attributes'):  # Missing if product was retrieved by id
            if stock_data := self.extension_attributes.get('stock_item', {}):
                return stock_data
        # Use the SKU to refresh attributes with full product data
        self.refresh()
        return self.stock_item

    @property
    def stock_item_id(self) -> int:
        """Item id of the StockItem, used to :meth:`~.update_stock`"""
        if self.stock_item:
            return self.stock_item['item_id']

    @property
    def description(self) -> str:
        """Product description (as HTML)"""
        return self.custom_attributes.get('description', '')

    @property
    def special_price(self) -> float:
        """The current special (sale) price"""
        return self.custom_attributes.get('special_price')


class MediaEntry(Model):

    """Wraps a media gallery entry of a :class:`Product`"""

    MEDIA_TYPES = ['base', 'small', 'thumbnail', 'swatch']

    DOCUMENTATION = "https://adobe-commerce.redoc.ly/2.3.7-admin/tag/productsskumediaentryId"
    IDENTIFIER = "id"

    def __init__(self, product: Product, entry: dict):
        """Initialize a MediaEntry object for a :class:`Product`

        :param product: the :class:`Product` that the gallery entry is associated with
        :param entry: the json response data to use as the source data
        """
        super().__init__(
            data=entry,
            client=product.client,
            endpoint=f'products/{product.encoded_sku}/media'
        )
        self.product = product

    def __repr__(self):
        return f"<MediaEntry {self.id} for {self.product}: {self.label}>"

    def query_endpoint(self) -> None:
        """No search endpoint exists for media gallery entries"""
        return self.logger.info("There is no search interface for media gallery entries")

    @property
    def excluded_keys(self) -> List[str]:
        return []

    @property
    def is_enabled(self):
        return not self.disabled

    @property
    def is_thumbnail(self):
        return 'thumbnail' in self.types

    @cached_property
    def link(self):
        """Permalink to the image"""
        return self.client.store.active.base_media_url + 'catalog/product' + self.file

    def disable(self, scope: Optional[str] = None) -> bool:
        """Disables the MediaEntry on the given scope

        :param scope: the scope to send the request on; will use the :attr:`.Client.scope` if not provided
        """
        self.data['disabled'] = True
        return self.update(scope)

    def enable(self, scope: Optional[str] = None) -> bool:
        """Enables the MediaEntry on the given scope

        :param scope: the scope to send the request on; will use the :attr:`.Client.scope` if not provided
        """
        self.data['disabled'] = False
        return self.update(scope)

    def add_media_type(self, media_type: str, scope: Optional[str] = None) -> bool:
        """Add a media type to the MediaEntry on the given scope

        .. caution:: If the media type is already assigned to a different entry, it will be removed

        :param media_type: one of the :attr:`~.MEDIA_TYPES`
        :param scope: the scope to send the request on; will use the :attr:`.Client.scope` if not provided
        """
        if media_type in self.MEDIA_TYPES and media_type not in self.types:
            self.data['types'].append(media_type)
            return self.update(scope)

    def remove_media_type(self, media_type: str, scope: Optional[str] = None) -> bool:
        """Remove a media type from the MediaEntry on the given scope

        :param media_type: one of the :attr:`~MEDIA_TYPES`
        :param scope: the scope to send the request on; will use the :attr:`.Client.scope` if not provided
        """
        if media_type in self.types:
            self.data['types'].remove(media_type)
            return self.update(scope)

        self.logger.error(f'{media_type} is not currently assigned to {self}')
        return False

    def set_media_types(self, types: list, scope: Optional[str] = None) -> bool:
        """Set media types for the MediaEntry on the given scope

        :param types: a list containing all :attr:`~MEDIA_TYPES` to assign
        :param scope: the scope to send the request on; will use the :attr:`.Client.scope` if not provided
        """
        if not isinstance(types, list):
            raise TypeError('types must be a list')

        self.data['types'] = [t for t in types if t in self.MEDIA_TYPES]
        return self.update(scope)

    def set_position(self, position: int, scope: Optional[str] = None) -> bool:
        """Set the position of the MediaEntry on the given scope

        :param position: the position to change to
        :param scope: the scope to send the request on; will use the :attr:`.Client.scope` if not provided
        """
        if not isinstance(position, int):
            raise TypeError('position must be an int')

        self.data['position'] = position
        return self.update(scope)

    def set_alt_text(self, text: str, scope: Optional[str] = None) -> bool:
        """Set the alt text (``label``) of the MediaEntry on the given scope

        :param text: the alt text to use
        :param scope: the scope to send the request on; will use the :attr:`.Client.scope` if not provided
        """
        if not isinstance(text, str):
            raise TypeError('text must be a string')

        self.data['label'] = text
        return self.update(scope)

    def update(self, scope: Optional[str] = None) -> bool:
        """Uses the :attr:`~.data` dict to update the media entry

        .. note:: Some updates alter the data of other entries; if the update is successful, the
            associated :class:`Product` will be refreshed on the same scope to keep the data consistent

        .. tip:: If there's only 1 store view, the admin will also be updated

        :param scope: the scope to send the request on; will use the :attr:`.Client.scope` if not provided
        """
        if self.client.store.is_single_store:
            success = self._update_single_store()
        else:
            success = self._update(scope)

        self.refresh(scope)  # Get updated data if success or reset to accurate data if failed

        if success:
            self.product.refresh(scope)

        return success

    def _update_single_store(self):
        """Updates the MediaEntry data on the default store view and admin"""
        for scope in (None, 'all'):
            if not self._update(scope):
                return False  # Avoid updating admin if store update fails
        return True

    def _update(self, scope: Optional[str] = None) -> bool:
        url = self.data_endpoint(scope)
        response = self.client.put(url, payload={'entry': self.data})

        if response.ok and response.json() is True:
            self.logger.info(
                f"Updated {self} on scope {self.get_scope_name(scope)}"
            )
            return True
        else:
            self.logger.error(
                f"Failed to update {self} on scope {self.get_scope_name(scope)}"
            )
            return False


class ProductAttribute(Model):

    """Wrapper for the ``products/attributes`` endpoint"""

    DOCUMENTATION = "https://adobe-commerce.redoc.ly/2.3.7-admin/tag/productsattributes/"
    IDENTIFIER = "attribute_code"

    def __init__(self, data: dict, client: Client):
        """Initialize a ProductAttribute object using an API response from the ``products/attributes`` endpoint

        :param data: the API response from the ``products/attributes`` endpoint
        :param client: an initialized :class:`~.Client` object
        """
        super().__init__(
            data=data,
            client=client,
            endpoint='products/attributes',
            private_keys=True
        )

    def __repr__(self):
        return f"<Product Attribute: {self.attribute_code}>"

    @property
    def excluded_keys(self) -> List[str]:
        return ['options']

    @property
    def options(self):
        return self.unpack_attributes(self.__options, key='label')



from . import clients
from . import search
from . import models
from . import utils
from . import exceptions
import os

__version__ = "2.1.1b0"

Client = clients.Client
logger = utils.MagentoLogger(
    name=utils.MagentoLogger.PACKAGE_LOG_NAME,
    log_file=utils.MagentoLogger.PACKAGE_LOG_NAME + '.log',
    stdout_level='WARNING'  # Clients will log to console
)


def get_api(**kwargs) -> Client:
    """Initialize a :class:`~.Client` using credentials stored in environment variables

    Any valid :class:`~.Client` kwargs can be used in addition to and/or instead of environment variables

    **Usage**::

      import magento

      api = magento.get_api()

    :param kwargs: any valid kwargs for :class:`~.Client`
    :raises ValueError: if login credentials are missing
    """
    credentials = {
        'domain': kwargs.get('domain', os.getenv('MAGENTO_DOMAIN')),
        'username': kwargs.get('username', os.getenv('MAGENTO_USERNAME')),
        'password': kwargs.get('password', os.getenv('MAGENTO_PASSWORD')),
        'local': kwargs.get('local', False),
    }
    if bad_keys := [key for key in credentials if credentials[key] is None]:
        raise ValueError(f'Missing login credentials for {bad_keys}')
    else:
        return Client.from_dict(credentials)


logger.debug('Initialized MyMagento')



import os
import sys
import logging
import requests

from typing import Union, List, Type
from logging import Logger, FileHandler, StreamHandler, Handler


class ItemManager:

    def __init__(self, items=None):
        self.items = items if items else []

    def add(self, item):
        if item not in self.items:
            self.items.append(item)

    def get_attrs(self, attr):
        return [getattr(item, attr, 0) for item in self.items]

    def sum_attrs(self, attr):
        return sum(self.get_attrs(attr))


class LoggerUtils:
    """Utility class that simplifies access to logger handler info"""

    @staticmethod
    def get_handler_names(logger) -> List[str]:
        """Get all handler names"""
        return [handler.name for handler in logger.handlers]

    @staticmethod
    def get_stream_handlers(logger: Logger) -> List[Handler]:
        """Get all the StreamHandlers of the current logger (NOTE: StreamHandler subclasses excluded)"""
        return [handler for handler in logger.handlers if type(handler) == StreamHandler]

    @staticmethod
    def get_file_handlers(logger: Logger) -> List[FileHandler]:
        """Get all the FileHandlers of the current logger"""
        return [handler for handler in logger.handlers if isinstance(handler, FileHandler)]

    @staticmethod
    def get_log_files(logger: Logger) -> List[str]:
        """Get the log file paths from all FileHandlers of a logger"""
        return [handler.baseFilename for handler in LoggerUtils.get_file_handlers(logger)]

    @staticmethod
    def get_handler_by_log_file(logger: Logger, log_file: str) -> Union[FileHandler, List[FileHandler]]:
        """Returns the FileHandler logging to the specified file, given it exists"""
        handlers = [
            handler for handler in LoggerUtils.get_file_handlers(logger)
            if os.path.basename(handler.baseFilename) == log_file
        ]
        if handlers:
            if len(handlers) == 1:
                return handlers[0]
            return handlers

    @staticmethod
    def clear_handlers(logger: Logger) -> bool:
        for handler in list(logger.handlers):
            logger.removeHandler(handler)
        return logger.handlers == []

    @staticmethod
    def clear_stream_handlers(logger: Logger) -> bool:
        """Removes all StreamHandlers from a logger"""
        for handler in LoggerUtils.get_stream_handlers(logger):
            logger.removeHandler(handler)
        return LoggerUtils.get_stream_handlers(logger) == []

    @staticmethod
    def clear_file_handlers(logger: Logger) -> bool:
        """Removes all FileHandlers from a logger"""
        for handler in LoggerUtils.get_file_handlers(logger):
            logger.removeHandler(handler)
        return LoggerUtils.get_file_handlers(logger) == []

    @staticmethod
    def map_handlers_by_name(logger: Logger):
        """Map the handlers of a logger first by type, and then by their name

        FileHandlers are mapped to both their handlers and log file, while StreamHandlers are just mapped to the handler
        Handlers without a name will be skipped, because look at the method name (:
        """
        mapping = {
            'stream': {},
            'file': {}
        }
        for stream_handler in LoggerUtils.get_stream_handlers(logger):
            if stream_handler.name:
                mapping['stream'][stream_handler.name] = stream_handler

        for file_handler in LoggerUtils.get_file_handlers(logger):
            if file_handler.name:
                entry = mapping['file'].setdefault(file_handler.name, {})
                entry['handler'] = file_handler
                entry['file'] = file_handler.baseFilename

        return mapping


class MagentoLogger:
    """Logging class used within the package

    :cvar PREFIX:           hardcoded prefix to use in log messages
    :cvar PACKAGE_LOG_NAME: the default name for the package logger
    :cvar CLIENT_LOG_NAME:  the default format for the client logger name
    :cvar LOG_MESSAGE:      the default format for the message component of log messages.
                            (Use magento.logger.LOG_MESSAGE for easy access)
    :cvar FORMATTER:        the default logging format
    :type FORMATTER:        logging.Formatter
    :cvar HANDLER_NAME:      the default format for the names of handlers created by this package
    """

    PREFIX = "MyMagento"
    PACKAGE_LOG_NAME = "my-magento"
    CLIENT_LOG_NAME = "{domain}_{username}"
    HANDLER_NAME = '{}__{}__{}'.format(PREFIX, '{name}', '{stdout_level}')

    LOG_MESSAGE = "|[ {pfx} | {name} ]|:  {message}".format(
        pfx=PREFIX, name="{name}", message="{message}"
    )

    FORMATTER = logging.Formatter(
        fmt="%(asctime)s %(levelname)-5s  %(message)s",
        datefmt="%Y-%m-%d %H:%M:%S"
    )

    def __init__(self, name: str, log_file: str = None, stdout_level: Union[int, str] = 'INFO', log_requests: bool = True):

        """Initialize the logger

        Each Client object corresponds to a unique username/domain combination, which is used to attach it to its
        associated MagentoLogger and log file, allowing all activity across all endpoints to be tracked.
        A package logger exists as well, which logs all activity from the package.
        All log files have their log level set to DEBUG

        :param name: logger name
        :param log_file: log file name; default is {name}.log
        :param stdout_level: logging level for stdout logger; default is "INFO" (which is also logging.INFO and 10)
        :param log_requests: set to True to add logging from the requests package logger
        """
        self.name = name
        self.logger = None
        self.handler_name = None
        self.log_file = log_file if log_file else f'{self.name}.log'
        self.setup_logger(stdout_level, log_requests=log_requests)

    def setup_logger(self, stdout_level: Union[int, str] = 'INFO', log_requests: bool = True) -> bool:
        """Configures a logger and assigns it to the `logger` attribute.

        :param stdout_level: logging level to use for logging to console
        :param log_requests: set to True to add logs from the requests package (ie. API call logging)
        """
        logger = logging.getLogger(self.name)
        log_files = LoggerUtils.get_log_files(logger)
        handler_map = LoggerUtils.map_handlers_by_name(logger)

        self.handler_name = MagentoLogger.HANDLER_NAME.format(
            name=self.name, stdout_level=stdout_level
        )
        if self.handler_name in handler_map['stream'] and self.handler_name in handler_map['file']:
            if self.log_path in log_files:
                self.logger = logger  # Log levels and log files are correct
                return True

        if self.handler_name not in handler_map['stream']:
            if len(handler_map['stream']) > 0:
                self.clear_magento_handlers(logger, handler_type=StreamHandler)
            # Resetting ensures only the desired level is logged to console
            stdout_handler = StreamHandler(stream=sys.stdout)
            stdout_handler.setFormatter(MagentoLogger.FORMATTER)
            stdout_handler.name = self.handler_name
            stdout_handler.setLevel(stdout_level)
            logger.addHandler(stdout_handler)

        # Remove all FileHandlers created by this package (except handler for magento.log)
        if self.handler_name not in handler_map['file'] or self.log_path not in log_files:
            if len(handler_map['file']) > 0:
                self.clear_magento_file_handlers(logger)
            f_handler = FileHandler(self.log_file)
            f_handler.setFormatter(MagentoLogger.FORMATTER)
            f_handler.name = self.handler_name
            f_handler.setLevel("DEBUG")
            logger.addHandler(f_handler)

        if log_requests:
            f_handler = LoggerUtils.get_handler_by_log_file(logger, self.log_file)  # In case it wasn't just created
            MagentoLogger.add_request_logging(f_handler)

        if self.name != MagentoLogger.PACKAGE_LOG_NAME:  # All clients have the handler added to them
            pkg_handler = MagentoLogger.get_package_handler()   # For writing to {PACKAGE_LOG_NAME}.log
            logger.addHandler(pkg_handler)
            f_handler = LoggerUtils.get_handler_by_log_file(logger, self.log_file)
            MagentoLogger.add_request_logging(f_handler)

        if self.name != MagentoLogger.PACKAGE_LOG_NAME:
            pkg_handler = MagentoLogger.get_package_handler()
            logger.addHandler(pkg_handler)  # For writing to {PACKAGE_LOG_NAME}.log

        logger.setLevel(logging.DEBUG)
        self.logger = logger
        return True

    def format_msg(self, msg: str) -> str:
        """Formats the :attr:`~.LOG_MESSAGE` using the specified message"""
        return MagentoLogger.LOG_MESSAGE.format(
            name=self.name,
            message=msg
        )

    def debug(self, msg):
        """Formats the :attr:`~.LOG_MESSAGE` with the specified message, then logs it with Logger.debug()"""
        return self.logger.debug(
            self.format_msg(msg)
        )

    def info(self, msg):
        """Formats the :attr:`~.LOG_MESSAGE` with the specified message, then logs it with Logger.info()"""
        return self.logger.info(
            self.format_msg(msg)
        )

    def error(self, msg):
        """Formats the :attr:`~.LOG_MESSAGE` with the specified message, then logs it with Logger.error()"""
        return self.logger.error(
            self.format_msg(msg)
        )

    def warning(self, msg):
        """Formats the :attr:`~.LOG_MESSAGE` with the specified message, then logs it with Logger.warning()"""
        return self.logger.warning(
            self.format_msg(msg)
        )

    def critical(self, msg):
        """Formats the :attr:`~.LOG_MESSAGE` with the specified message, then logs it with Logger.critical()"""
        return self.logger.critical(
            self.format_msg(msg)
        )

    @property
    def handlers(self):
        return self.logger.handlers

    @property
    def handler_names(self):
        return LoggerUtils.get_handler_names(self.logger)

    @property
    def handler_map(self):
        return LoggerUtils.map_handlers_by_name(self.logger)

    @property
    def file_handlers(self):
        return LoggerUtils.get_file_handlers(self.logger)

    @property
    def stream_handlers(self):
        return LoggerUtils.get_stream_handlers(self.logger)

    @property
    def log_files(self):
        return LoggerUtils.get_log_files(self.logger)

    @property
    def log_path(self):
        return os.path.abspath(self.log_file)

    @staticmethod
    def get_magento_handlers(logger):
        return [handler for handler in logger.handlers if MagentoLogger.owns_handler(handler)]

    @staticmethod
    def clear_magento_handlers(logger: Logger, handler_type: Union[Type[FileHandler], Type[StreamHandler]], clear_pkg: bool = False) -> None:
        """Clear all handlers from a logger that were created by MagentoLogger

        :param logger: any logger
        :param handler_type: the logging handler type to check for and remove
        :param clear_pkg: if True, will delete the package handler for writing to my-magento.log (Default is False)
        """
        for handler in MagentoLogger.get_magento_handlers(logger):
            if type(handler) == handler_type:
                if clear_pkg is True or handler != MagentoLogger.get_package_handler():
                    logger.removeHandler(handler)  # Either remove all handlers, or all but pkg handler

    @staticmethod
    def clear_magento_file_handlers(logger: Logger, clear_pkg: bool = False):
        return MagentoLogger.clear_magento_handlers(logger, FileHandler, clear_pkg)

    @staticmethod
    def clear_magento_stdout_handlers(logger: Logger, clear_pkg: bool = False):
        return MagentoLogger.clear_magento_handlers(logger, StreamHandler, clear_pkg)

    @staticmethod
    def owns_handler(handler: Handler):
        """Checks if a handler is a Stream/FileHandler from this package or not"""
        try:  # Match handler name to MagentoLogger.HANDLER_NAME format
            prefix, name, stdout_level = handler.name.split('__')
            return prefix == MagentoLogger.PREFIX
        except:  # Wrong format or not set
            return False

    @staticmethod
    def get_package_handler() -> FileHandler:
        """Returns the FileHandler object that writes to the magento.log file"""
        pkg_handlers = logging.getLogger(MagentoLogger.PACKAGE_LOG_NAME).handlers
        for handler in pkg_handlers:
            if isinstance(handler, FileHandler):
                if handler.baseFilename == os.path.abspath(MagentoLogger.PACKAGE_LOG_NAME + '.log'):
                    return handler

    @staticmethod
    def add_request_logging(handler: Union[FileHandler, StreamHandler]):
        """Adds the specified handler to the requests package logger, allowing for easier debugging of API calls"""
        if type(handler) not in (FileHandler, StreamHandler):
            raise TypeError(f"Parameter handler must be of type {FileHandler} or {StreamHandler}")

        req_logger = requests.urllib3.connectionpool.log
        req_logger.setLevel("DEBUG")
        if handler in req_logger.handlers:
            return True  # Already added

        if type(handler) is FileHandler:
            if handler.baseFilename not in LoggerUtils.get_log_files(req_logger):
                req_logger.addHandler(handler)  # Might be same handler new file (or level)

        elif type(handler) is StreamHandler:
            stdout_names = LoggerUtils.map_handlers_by_name(req_logger)['stream']
            if handler.name not in stdout_names: # Might be same handler new level
                req_logger.addHandler(handler)

        return True


AGENTS = ['Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/102.0.5005.63 Safari/537.36']


def get_agents() -> list:
    """Scrapes a list of user agents. Returns a default list if the scrape fails."""
    if (response := requests.get('https://www.whatismybrowser.com/guides/the-latest-user-agent/chrome')).ok:
        section = response.text.split('<h2>Latest Chrome on Windows 10 User Agents</h2>')[1]
        raw_agents = section.split('code\">')[1:]
        agents = [agent.split('<')[0] for agent in raw_agents]
        for a in agents:
            if a not in AGENTS:
                AGENTS.append(a)
    # If function fails, will still return the hardcoded list
    return AGENTS


def get_agent(index=0) -> str:
    """Returns a single user agent string from the specified index of the AGENTS list"""
    return get_agents()[index]  # Specify index only if you hardcode more than 1


def get_package_file_handler():
    return MagentoLogger.get_package_handler()



from __future__ import annotations
from . import Model, Order, OrderItem, Product
from typing import TYPE_CHECKING, Optional, List, Union
from functools import cached_property

if TYPE_CHECKING:
    from magento import Client
    from magento.search import SearchQuery


class Invoice(Model):

    """Wrapper for the ``invoices`` endpoint"""

    DOCUMENTATION = 'https://adobe-commerce.redoc.ly/2.3.7-admin/tag/invoices'
    IDENTIFIER = 'entity_id'

    def __init__(self, data: dict, client: Client):
        """Initialize an Invoice object using an API response from the ``invoices`` endpoint

        :param data: API response from the ``invoices`` endpoint
        :param client: an initialized :class:`~.Client` object
        """
        super().__init__(
            data=data,
            client=client,
            endpoint='invoices',
            private_keys=True
        )

    def __repr__(self):
        return f'<Magento Invoice: #{self.number}> for {self.order}'

    @property
    def excluded_keys(self) -> List[str]:
        return ['items']

    @property
    def id(self) -> int:
        """Alias for ``entity_id``"""
        return getattr(self, 'entity_id', None)

    @property
    def number(self) -> str:
        """Alias for ``increment_id``"""
        return getattr(self, 'increment_id', None)

    @cached_property
    def order(self) -> Order:
        """The corresponding :class:`~.Order`"""
        return self.client.orders.by_id(self.order_id)

    @cached_property
    def items(self) -> List[InvoiceItem]:
        """The invoiced items, returned as a list of :class:`InvoiceItem` objects"""
        return [InvoiceItem(item, self) for item in self.__items
                if item['order_item_id'] in self.order.item_ids]


class InvoiceItem(Model):

    """Wraps an item entry of an :class:`Invoice`"""

    DOCUMENTATION = "https://adobe-commerce.redoc.ly/2.3.7-admin/tag/invoicesid"
    IDENTIFIER = 'entity_id'

    def __init__(self, item: dict, invoice: Invoice):
        """Initialize an InvoiceItem of an :class:`Invoice`

        :param item: API response to use as source data
        :param invoice: the :class:`Invoice` that this is an item of
        """
        super().__init__(
            data=item,
            client=invoice.client,
            endpoint=f'invoices/{invoice.id}'
        )
        self.invoice = invoice

    def __repr__(self):
        return f"<InvoiceItem ({self.sku})> from {self.invoice}"

    def data_endpoint(self, scope: Optional[str] = None) -> None:
        """No data endpoint exists for invoice items"""
        return self.logger.info("There is no API endpoint for individual invoice items")

    def query_endpoint(self) -> None:
        """No search endpoint exists for invoice items"""
        return self.logger.info("There is no API endpoint for querying invoice items")

    @property
    def excluded_keys(self) -> List[str]:
        return ['product_id']

    @property
    def order(self) -> Order:
        """The :class:`~.Order` this item is from"""
        return self.invoice.order

    @cached_property
    def order_item(self) -> OrderItem:
        """The item's corresponding :class:`~.OrderItem`"""
        for item in self.order.items:
            if item.item_id == self.order_item_id:
                return item

    @property
    def product(self) -> Product:
        """The item's corresponding simple :class:`~.Product`"""
        return self.order_item.product

    @property
    def product_id(self) -> int:
        """Id of the corresponding simple :class:`~.Product`"""
        return self.order_item.product_id



from __future__ import annotations
import copy
from . import Model
from typing import TYPE_CHECKING, List, Optional, Set, Dict, Union
from functools import cached_property

if TYPE_CHECKING:
    from magento import Client
    from . import Product, Order, OrderItem, Invoice


class Category(Model):

    """Wrapper for the ``categories`` endpoint"""

    DOCUMENTATION = 'https://adobe-commerce.redoc.ly/2.3.7-admin/tag/categories'
    IDENTIFIER = 'id'

    def __init__(self, data: dict, client: Client):
        """Initialize a Category object using an API response from the ``categories`` endpoint

        :param data: raw API response
        """
        super().__init__(
            data=data,
            client=client,
            endpoint='categories'
        )

    def __repr__(self):
        return f'<Magento Category: {self.name}>'

    @property
    def excluded_keys(self):
        return ['custom_attributes']

    @cached_property
    def custom_attributes(self) -> Dict[str, str]:
        return self.unpack_attributes(self.data.get('custom_attributes'))

    @cached_property
    def subcategories(self) -> List[Category]:
        """The child categories, returned as a list of :class:`~.Category` objects

        .. note:: Only the direct child categories are returned.
           For a list of all descendants, use :attr:`~.all_subcategories`
        """
        if hasattr(self, 'children_data'):  # A list of API response dicts (when Category is from search)
            return [self.parse(child) for child in self.children_data]
        if hasattr(self, 'children') and self.children:  # String of subcategory ids (from categories/{id} endpoint)
            return self.query_endpoint().by_list('entity_id', self.children)
        else:
            return []

    @cached_property
    def subcategory_ids(self) -> List[int]:
        """The ``category_ids`` of the :attr:`~.subcategories`"""
        if hasattr(self, 'children'):
            if self.children:
                return [int(child) for child in self.children.split(',')]
            return []
        return [category.id for category in self.subcategories]

    @cached_property
    def subcategory_names(self) -> List[str]:
        """The names of the category's :attr:`~.subcategories`"""
        return [category.name for category in self.subcategories]

    @cached_property
    def all_subcategories(self) -> Optional[List[Category]]:
        """Recursively retrieves all descendants of the category"""
        if not self.subcategories:
            return []
        children = copy.deepcopy(self.subcategories)
        for child in self.subcategories:
            children.extend(child.all_subcategories)
        return children

    @cached_property
    def all_subcategory_ids(self) -> List[int]:
        """The ``category_ids`` of :attr:`~.all_subcategories`"""
        return [category.id for category in self.all_subcategories]

    @cached_property
    def products(self) -> List[Product]:
        """The :class:`~.Product` s in the category

        Alias for :meth:`get_products`
        """
        return self.get_products()

    @cached_property
    def product_ids(self) -> List[int]:
        """The ``product_ids`` of the category's :attr:`~.products`"""
        return [product.id for product in self.products]

    @cached_property
    def skus(self) -> List[str]:
        """The skus of the category's :attr:`~.products`"""
        return [product.sku for product in self.products]

    @cached_property
    def all_products(self) -> List[Product]:
        """The :class:`~.Product` s in the category and in :attr:`~.all_subcategories`

        Alias for :meth:`get_products` with ``search_subcategories=True``
        """
        return self.get_products(search_subcategories=True)

    @cached_property
    def all_product_ids(self) -> Set[int]:
        """The ``product_ids`` of the products in the category and in :attr:`~.all_subcategories`"""
        return set(product.id for product in self.all_products)

    @cached_property
    def all_skus(self) -> Set[str]:
        """The skus of the products in the category and in :attr:`~.all_subcategories`"""
        return set(product.sku for product in self.all_products)

    def get_products(self, search_subcategories: bool = False) -> Optional[Product | List[Product]]:
        """Retrieves the category's products

        :param search_subcategories: if ``True``, also retrieves products from :attr:`~.all_subcategories`
        """
        return self.client.products.by_category(self, search_subcategories) or []

    def get_orders(self, search_subcategories: bool = False) -> Optional[Order | List[Order]]:
        """Retrieve any :class:`~.Order` that contains one of the category's :attr:`~products`

        :param search_subcategories: if ``True``, also searches for orders from :attr:`~.all_subcategories`
        """
        return self.client.orders.by_category(self, search_subcategories)

    def get_order_items(self, search_subcategories: bool = False) -> Optional[OrderItem | List[OrderItem]]:
        """Retrieve any :class:`~.OrderItem` that contains one of the category's :attr:`~products`

        :param search_subcategories: if ``True``, also searches for order items from :attr:`~.all_subcategories`
        """
        return self.client.order_items.by_category(self, search_subcategories)

    def get_invoices(self, search_subcategories: bool = False) -> Optional[Invoice | List[Invoice]]:
        """Retrieve any :class:`~.Invoice` that contains one of the category's :attr:`~products`

        :param search_subcategories: if ``True``, also searches for invoices from :attr:`~.all_subcategories`
        """
        return self.client.invoices.by_category(self, search_subcategories)



import os
from pathlib import Path
from setuptools import setup, find_packages


LONG_DESCRIPTION_SRC = 'README_PyPi.rst'


def read(file):
    with open(os.path.abspath(file), 'r', encoding='utf-8') as f:
        return f.read()


# Parse version
init = Path(__file__).parent.joinpath("magento", "__init__.py")
for line in init.read_text().split("\n"):
    if line.startswith("__version__ ="):
        break
version = line.split(" = ")[-1].strip('"')

setup(
    name='my-magento',
    packages=find_packages(),
    version=version,
    license='MIT',
    description='Python Magento 2 REST API Wrapper',
    long_description=read(LONG_DESCRIPTION_SRC),
    long_description_content_type="text/x-rst; charset=UTF-8",
    author='Adam Korn',
    author_email='hello@dailykitten.net',
    url='https://www.github.com/TDKorn/my-magento',
    download_url="https://github.com/TDKorn/my-magento/tarball/master",
    keywords=["magento", "magento-api", "python-magento", "python", "python3", "magento-python", "pymagento", "py-magento", "magento2", "magento-2", "magento2-api"],
    install_requires=["requests"]
)
